import unittest
import itertools
import ir_measures


class TestPytrecEval(unittest.TestCase):

    def test_nDCG(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 2
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.gdeval
        measure = ir_measures.nDCG@20
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.6201)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.35099)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.485545)
        self.assertEqual(provider.evaluator([measure], qrels).calc_aggregate(run)[measure], 0.485545)

        measure = ir_measures.nDCG@2
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.17377)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.38685)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.28031)

        ev = provider.evaluator([ir_measures.nDCG@20, ir_measures.nDCG@2], qrels)
        res = ev.calc_aggregate(run)
        self.assertEqual(res[ir_measures.nDCG@20], 0.485545)
        self.assertEqual(res[ir_measures.nDCG@2], 0.28031)
        res = ev.calc_aggregate(run)
        self.assertEqual(res[ir_measures.nDCG@20], 0.485545)
        self.assertEqual(res[ir_measures.nDCG@2], 0.28031)

    def test_ERR(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 2
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.gdeval
        measure = ir_measures.ERR@20
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.10175)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.09375)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.09775)

        measure = ir_measures.ERR@2
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.03125)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.09375)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.0625)



if __name__ == '__main__':
    unittest.main()



from ir_measures import measures


class _Bpref(measures.Measure):
    """
    Binary Preference (Bpref).
    This measure examines the relative ranks of judged relevant and non-relevant documents. Non-judged documents are not considered. 

    .. cite.dblp:: conf/sigir/BuckleyV04
    """
    __name__ = 'Bpref'
    NAME = __name__
    PRETTY_NAME = 'Binary Preference'
    SHORT_DESC = 'The relative ranks of judged relevant and non-relevant documents.'
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)')
    }


Bpref = _Bpref()
BPref = Bpref
measures.register(Bpref, ['BPref'])



from ir_measures.providers.base import Provider, Evaluator, registry, register
from ir_measures.providers.accuracy_provider import AccuracyProvider
from ir_measures.providers.fallback_provider import FallbackProvider
from ir_measures.providers.compat_provider import CompatProvider
from ir_measures.providers.cwl_eval import CwlEvalProvider, CwlMetric
from ir_measures.providers.pyndeval_provider import PyNdEvalProvider
from ir_measures.providers.pytrec_eval_provider import PytrecEvalProvider
from ir_measures.providers.judged_provider import JudgedProvider
from ir_measures.providers.gdeval_provider import GdevalProvider
from ir_measures.providers.trectools_provider import TrectoolsProvider
from ir_measures.providers.msmarco_provider import MsMarcoProvider
from ir_measures.providers.ranx_provider import RanxProvider
from ir_measures.providers.runtime_provider import RuntimeProvider, define, define_byquery

__all__ = [
	'registry', 'register',
	'Provider', 'Evaluator',
	'AccuracyProvider', 'FallbackProvider', 'CompatProvider', 'CwlEvalProvider', 'CwlMetric', 'PyNdEvalProvider',
	'PytrecEvalProvider', 'JudgedProvider', 'GdevalProvider', 'TrectoolsProvider', 'MsMarcoProvider',
	'RanxProvider', 'RuntimeProvider',
	'define', 'define_byquery',
]



import unittest
import itertools
import ir_measures
import pandas as pd

class TestRuntime(unittest.TestCase):

    def test_non_doc_qrels(self):
        # source: https://thecleverprogrammer.com/2022/07/27/longest-common-prefix-using-python/
        def longestCommonPrefix(strs):
            l = len(strs[0])
            for i in range(1, len(strs)):
                length = min(l, len(strs[i]))
                while length > 0 and strs[0][0:length] != strs[i][0:length]:
                    length = length - 1
                    if length == 0:
                        return 0
            return strs[0][0:length]

        LenMeasure = ir_measures.define_byquery(lambda qrels, run: len(run.iloc[0]["qanswer"]))
        LCSMeasure = ir_measures.define_byquery(lambda qrels, run: len(longestCommonPrefix(
            [qrels.iloc[0]["gold_answer"],
            run.iloc[0]["qanswer"]]
        )))
        
        prefix = 'professor proton mixed the '
        test_answer = prefix + 'reactants'
        gold_answer = prefix + 'chemicals'
        df_res = pd.DataFrame([['q1', test_answer]], columns=['query_id', 'qanswer'])
        df_gold = pd.DataFrame([['q1', gold_answer]], columns=['query_id', 'gold_answer'])
        measurements = ir_measures.iter_calc([LenMeasure, LCSMeasure], df_gold, df_res)
        for m in measurements:
            if m.measure == LenMeasure:
                self.assertEqual(len(test_answer), m.value)
            elif m.measure == LCSMeasure:
                self.assertEqual(len(prefix), m.value)
            else:
                raise Exception()

    def test_define_byquery(self):
        def my_p(qrels, run):
            run = run.merge(qrels, 'left', on=['doc_id'])
            return (run['relevance'] > 0).sum() / len(run)
        def my_s(qrels, run):
            run = run.merge(qrels, 'left', on=['doc_id'])
            return 1. if (run['relevance'] > 0).sum() else 0.
        MyP = ir_measures.define_byquery(my_p)
        MyS = ir_measures.define_byquery(my_s)
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 0
1 0 D5 2
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))

        result = list((MyP@1).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyP@1).calc_aggregate(qrels, run), 0.0)

        result = list((MyP@2).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.5)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyP@2).calc_aggregate(qrels, run), 0.25)

        result = list((MyP@3).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.6666666666666666)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyP@3).calc_aggregate(qrels, run), 0.3333333333333333)

        result = list((MyS@2).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1.)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyS@2).calc_aggregate(qrels, run), 0.5)

    def test_define(self):
        def my_p(qrels, run):
            run = run.merge(qrels, 'left', on=['query_id', 'doc_id'])
            for qid, df in run.groupby('query_id'):
                yield qid, (df['relevance'] > 0).sum() / len(df)
        def my_s(qrels, run):
            run = run.merge(qrels, 'left', on=['query_id', 'doc_id'])
            for qid, df in run.groupby('query_id'):
                yield qid, 1. if (df['relevance'] > 0).sum() else 0.
        MyP = ir_measures.define(my_p)
        MyS = ir_measures.define(my_s)
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 0
1 0 D5 2
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))
        result = list((MyP@1).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyP@1).calc_aggregate(qrels, run), 0.0)

        result = list((MyP@2).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.5)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyP@2).calc_aggregate(qrels, run), 0.25)

        result = list((MyP@3).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.6666666666666666)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyP@3).calc_aggregate(qrels, run), 0.3333333333333333)

        result = list((MyS@2).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1.)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyS@2).calc_aggregate(qrels, run), 0.5)

        # check empty lists
        result = list((MyS@2).iter_calc(qrels, []))
        result = list((MyS@2).iter_calc([], run))
        result = list((MyS@2).iter_calc([], []))

    def test_run_qrel_inputs(self):
        def a(qrels, run):
            ...
        def b(qrels, run):
            ...
        A = ir_measures.define(a, run_inputs=['query_id'], qrel_inputs=['query_id', 'doc_id', 'x'])
        B = ir_measures.define(b)
        provider = ir_measures.providers.RuntimeProvider()
        inputs = provider.run_inputs([A, B])
        self.assertEqual(set(inputs), {'query_id', 'doc_id', 'score'})
        inputs = provider.run_inputs([A])
        self.assertEqual(set(inputs), {'query_id'})
        inputs = provider.run_inputs([B])
        self.assertEqual(set(inputs), {'query_id', 'doc_id', 'score'})
        inputs = ir_measures.DefaultPipeline.run_inputs([A, B])
        self.assertEqual(set(inputs), {'query_id', 'doc_id', 'score'})
        inputs = ir_measures.DefaultPipeline.run_inputs([A])
        self.assertEqual(set(inputs), {'query_id'})
        inputs = ir_measures.DefaultPipeline.run_inputs([A, ir_measures.P@5])
        self.assertEqual(set(inputs), {'query_id', 'doc_id', 'score'})

        inputs = provider.qrel_inputs([A, B])
        self.assertEqual(set(inputs), {'query_id', 'doc_id', 'relevance', 'x'})
        inputs = provider.qrel_inputs([A])
        self.assertEqual(set(inputs), {'query_id', 'doc_id', 'x'})
        inputs = provider.qrel_inputs([B])
        self.assertEqual(set(inputs), {'query_id', 'doc_id', 'relevance'})
        inputs = ir_measures.DefaultPipeline.qrel_inputs([A, B])
        self.assertEqual(set(inputs), {'query_id', 'doc_id', 'relevance', 'x'})
        inputs = ir_measures.DefaultPipeline.qrel_inputs([A])
        self.assertEqual(set(inputs), {'query_id', 'doc_id', 'x'})
        inputs = ir_measures.DefaultPipeline.qrel_inputs([A, ir_measures.P@5])
        self.assertEqual(set(inputs), {'query_id', 'doc_id', 'relevance', 'x'})



if __name__ == '__main__':
    unittest.main()



from ir_measures import measures


class _IPrec(measures.Measure):
    """
    Interpolated Precision at a given recall cutoff. Used for building precision-recall graphs.
    Unlike most measures, where @ indicates an absolute cutoff threshold, here @ sets the recall
    cutoff.
    """
    __name__ = 'IPrec'
    NAME = __name__
    PRETTY_NAME = 'Interpolated Precision@recall'
    SHORT_DESC = 'The interpolated precision at a given recall cutoff.'
    AT_PARAM = 'recall'
    SUPPORTED_PARAMS = {
        'recall': measures.ParamInfo(dtype=float, required=True, desc='recall threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }


IPrec = _IPrec()
measures.register(IPrec)



from itertools import zip_longest
import unittest
import ir_measures
from ir_measures.measures import Accuracy

class TestMeasures(unittest.TestCase):

    def test_accuracy(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 C0_1 0
0 0 B0_1 1
0 0 A0_1 2

1 0 A1_1 2
1 0 A1_2 2
1 0 B1_1 1
1 0 C1_2 0

2 0 B1_1 1
'''))
        run = list(ir_measures.read_trec_run('''
0 0 C0_1 1 0.4 run
0 0 A0_1 2 0.3 run
0 0 C0_2 3 0.2 run
0 0 C0_3 4 0.1 run

1 0 C1_1 1 0.8 run
1 0 A1_1 2 0.7 run
1 0 C1_2 3 0.6 run
1 0 B1_1 4 0.5 run
1 0 C1_3 5 0.4 run

2 0 B1_1 2 0.2 run 
2 0 C1_1 3 0.1 run
'''))
        provider = ir_measures.accuracy

        accuracy_1 = Accuracy(rel=1)
        results_1 = [('0', 2./3.), ('1', .5 * (2/3. + 1./3)), ('2', 1.)]
        expected_results = [
            [accuracy_1, results_1],
            [Accuracy(rel=2), [('0', 2./3), ('1', 0.75)]],
        ]
        for measure, expected in expected_results:
            with self.subTest(measure=measure):
                self.assertTrue(provider.supports(measure))
                results = list(provider.iter_calc([measure], qrels, run))
                self.assertEqual(len(results), len(expected), "result lists length differ")

                for result, (query_id, value) in zip(results, expected):
                    self.assertEqual(result.query_id, query_id)
                    self.assertAlmostEqual(result.value, value, delta=1e-9, msg=f"for query {query_id}")

        expected = sum(value for _, value in results_1) / len(results_1)
        self.assertAlmostEqual(provider.calc_aggregate([accuracy_1], qrels, run)[accuracy_1], expected, delta=1e-9)


if __name__ == '__main__':
    unittest.main()



import ir_measures
from ir_measures import providers, measures, Metric
from ir_measures.providers.base import Any, Choices, NOT_PROVIDED


class RanxProvider(providers.Provider):
    """
    ranx

    https://amenra.github.io/ranx/


    .. cite.dblp:: conf/ecir/Bassani22
    """
    NAME = 'ranx'
    SUPPORTED_MEASURES = [
        measures._P(cutoff=Any(), rel=Any(), judged_only=Choices(False)),
        measures._SetP(rel=Any(), judged_only=Choices(False)),
        measures._RR(cutoff=Choices(NOT_PROVIDED), rel=Any(), judged_only=Choices(False)),
        measures._Rprec(rel=Any(), judged_only=Choices(False)),
        measures._AP(cutoff=Any(), rel=Any(), judged_only=Choices(False)),
        measures._nDCG(cutoff=Any(), dcg=Choices('log2', 'exp-log2'), gains=Choices(NOT_PROVIDED), judged_only=Choices(False)),
        measures._R(cutoff=Any(), judged_only=Choices(False)),
        measures._SetR(rel=Any()),
        measures._NumRet(rel=Any(required=True)),
        measures._Success(cutoff=Any(required=True), rel=Any(), judged_only=Choices(False)),
    ]

    def __init__(self):
        super().__init__()
        self.ranx = None

    def _evaluator(self, measures, qrels):
        # Convert qrels to dict_of_dict (input format used by pytrec_eval)
        qrels = ir_measures.util.QrelsConverter(qrels).as_pd_dataframe()
        qids = set(qrels['query_id'].unique())

        # Depending on the measure params, we may need multiple invocations of ranx
        # (e.g., with different rel_level, since it only supports running with 1 rel_level at a time)
        invokers = self._build_invokers(measures, qrels)
        return RanxEvaluator(self.ranx, measures, invokers, qrels, qids=qids)

    def _build_invokers(self, measures, qrels):
        invocations = {}
        for measure in measures:
            match_str = None
            if measure.NAME == 'P':
                invocation_key = (measure['rel'], 0)
                measure_str = f'precision@{measure["cutoff"]}'
            elif measure.NAME == 'SetP':
                invocation_key = (measure['rel'], 0)
                measure_str = 'precision'
            elif measure.NAME == 'R':
                invocation_key = (measure['rel'], 0)
                measure_str = f'recall@{measure["cutoff"]}'
            elif measure.NAME == 'SetR':
                invocation_key = (measure['rel'], 0)
                measure_str = 'recall'
            elif measure.NAME == 'RR':
                invocation_key = (measure['rel'], 0)
                if 'cutoff' in measure.params:
                    measure_str = f'mrr@{measure["cutoff"]}'
                else:
                    measure_str = 'mrr'
            elif measure.NAME == 'AP':
                invocation_key = (measure['rel'], 0)
                if 'cutoff' in measure.params:
                    measure_str = f'map@{measure["cutoff"]}'
                else:
                    measure_str = 'map'
            elif measure.NAME == 'Success':
                invocation_key = (measure['rel'], 0)
                measure_str = f'hit_rate@{measure["cutoff"]}'
            elif measure.NAME == 'NumRet':
                invocation_key = (measure['rel'], 0)
                measure_str = 'hits'
            elif measure.NAME == 'nDCG':
                invocation_key = (None, 0)
                name = 'ndcg_burges' if measure.params.get('dcg', measure.SUPPORTED_PARAMS['dcg'].default) == 'exp-log2' else 'ndcg'
                if 'cutoff' in measure.params:
                    measure_str = f'{name}@{measure["cutoff"]}'
                else:
                    measure_str = name
            elif measure.NAME == 'Rprec':
                invocation_key = (measure['rel'], 0)
                measure_str = 'r-precision'
            else:
                raise ValueError(f'unsupported measure {measure}')

            if match_str is None:
                match_str = measure_str

            if invocation_key not in invocations:
                invocations[invocation_key] = {}
            invocations[invocation_key][match_str] = (measure, measure_str)

        invokers = []
        for (rel_level, it), measure_map in invocations.items():
            if rel_level is not None:
                these_qrels = qrels.assign(relevance=(qrels['relevance']>=rel_level).astype(int))
            else:
                these_qrels = qrels
            these_qrels = self.ranx.Qrels.from_df(these_qrels, q_id_col='query_id', doc_id_col='doc_id', score_col='relevance')
            invokers.append(RanxInvoker(self.ranx, these_qrels, measure_map))

        return invokers

    def initialize(self):
        try:
            import ranx
            self.ranx = ranx
        except ImportError as ex:
            raise RuntimeError('ranx not available (do you need to `pip install ranx`?)', ex)

    def install_instructions(self):
        return 'pip install ir-measures[ranx]'


class RanxEvaluator(providers.Evaluator):
    def __init__(self, ranx, measures, invokers, qrels, qids):
        super().__init__(measures, qids)
        self.ranx = ranx
        self.invokers = invokers

    def _iter_calc(self, run):
        run = self.ranx.Run.from_df(ir_measures.util.RunConverter(run).as_pd_dataframe(), q_id_col='query_id', doc_id_col='doc_id', score_col='score')
        for invoker in self.invokers:
            yield from invoker.iter_calc(run)


class RanxInvoker:
    def __init__(self, ranx, qrels, measure_map):
        self.ranx = ranx
        self.qrels = qrels
        self.measure_map = measure_map

    def iter_calc(self, run):
        self.ranx.evaluate(self.qrels, run, list(self.measure_map))
        for measure, qid_value_map in run.scores.items():
            for query_id, value in qid_value_map.items():
                yield Metric(query_id=query_id, measure=self.measure_map[measure][0], value=value)
        run.scores.clear()


providers.register(RanxProvider())



from ir_measures import measures


class _Rprec(measures.Measure):
    """
    The precision at R, where R is the number of relevant documents for a given query. Has the cute property that
    it is also the recall at R.

    .. cite:: retrieval-system-evaluation
        :citation: Buckley and Voorhees. Retrieval System Evaluation. 2005.
        :link: https://www.nist.gov/publications/retrieval-system-evaluation
    """
    __name__ = 'Rprec'
    NAME = __name__
    PRETTY_NAME = 'Precsion at R'
    SHORT_DESC = 'Precsion at R, where R is the number of relevant documents for a given query.'
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }


Rprec = _Rprec()
RPrec = Rprec
measures.register(Rprec, ['RPrec'])



import warnings
import re
import io
import ast
import gzip
import contextlib
import itertools
import tempfile
from typing import Dict, List
from collections import defaultdict
from typing import NamedTuple, Union, Iterable
import ir_measures
if False: # this is to allow type-checking for pandas
    import pandas

class Qrel(NamedTuple):
    query_id: str
    doc_id: str
    relevance: int
    iteration: str = '0'

class ScoredDoc(NamedTuple):
    query_id: str
    doc_id: str
    score: float

class Metric(NamedTuple):
    query_id: str
    measure: 'ir_measures.Measure'
    value: Union[float, int]

class CalcResults(NamedTuple):
    aggregated: Union[Dict['ir_measures.Measure', Union[float, int]], float, int]
    per_query: List[Metric]

class GenericQrel(Qrel):
    def __new__(self, *args, **kwargs):
        warnings.warn("GenericQrel deprecated in 0.2.0. Please use ir_measures.Qrel instead.", DeprecationWarning)
        return super().__new__(self, *args, **kwargs)
GenericQrel._fields = Qrel._fields


class GenericScoredDoc(ScoredDoc):
    def __new__(self, *args, **kwargs):
        warnings.warn("GenericScoredDoc deprecated in 0.2.0. Please use ir_measures.ScoredDoc instead.", DeprecationWarning)
        return super().__new__(self, *args, **kwargs)
GenericScoredDoc._fields = ScoredDoc._fields

TYPE_RUN = Union[ Iterable[ScoredDoc], 'pandas.DataFrame', Dict[str, Dict[str, float]]]
TYPE_QREL = Union[ Iterable[Qrel], 'pandas.DataFrame', Dict[str, Dict[str, int]]] 

class QrelsConverter:
    def __init__(self, qrels, strict=True):
        self.qrels = qrels
        self._predicted_format = None
        self.strict = strict # setting strict to false prevents missing columns from raising an error for DFs

    def tee(self, count):
        t, err = self.predict_type()
        if t == 'namedtuple_iter':
            teed_qrels = itertools.tee(self.qrels, count)
            return [QrelsConverter(qrels) for qrels in teed_qrels]
        return [self for _ in range(count)]

    def predict_type(self):
        if self._predicted_format:
            return self._predicted_format
        result = 'UNKNOWN'
        error = None
        if isinstance(self.qrels, dict):
            result = 'dict_of_dict'
        elif hasattr(self.qrels, 'itertuples'):
            cols = self.qrels.columns
            missing_cols = [f for f in Qrel._fields if f not in cols and f not in Qrel._field_defaults]
            if missing_cols and self.strict:
                error = f'DataFrame missing columns: {list(missing_cols)} (found {list(cols)})'
            else:
                result = 'pd_dataframe'                
        elif hasattr(self.qrels, '__iter__'):
            # peek
            # TODO: is this an OK approach?
            self.qrels, peek_qrels = itertools.tee(self.qrels, 2)
            sentinal = object()
            item = next(peek_qrels, sentinal)
            if isinstance(item, tuple) and hasattr(item, '_fields'):
                fields = item._fields
                missing_fields = [f for f in Qrel._fields if f not in fields and f not in Qrel._field_defaults]
                if not missing_fields:
                    result = 'namedtuple_iter'
                else:
                    error = f'namedtuple iter missing fields: {list(missing_fields)} (found {list(fields)})'
            elif item is sentinal:
                result = 'namedtuple_iter'
            else:
                error = 'iterable not a namedtuple iterator'
        else:
            required_fields = tuple(f for f in Qrel._fields if f not in Qrel._field_defaults)
            error = f'unexpected format; please provide either: (1) an iterable of namedtuples (fields {required_fields}, e.g., from ir_measures.Qrel); (2) a pandas DataFrame with columns {required_fields}; or (3) a dict-of-dict'
        self._predicted_format = (result, error)
        return result, error

    def as_dict_of_dict(self):
        t, err = self.predict_type()
        if t == 'dict_of_dict':
            return self.qrels
        else:
            result = {}
            for qrel in self.as_namedtuple_iter():
                if qrel.query_id not in result:
                    result[qrel.query_id] = {}
                result[qrel.query_id][qrel.doc_id] = qrel.relevance
            return result

    def as_namedtuple_iter(self):
        t, err = self.predict_type()
        if t == 'namedtuple_iter':
            yield from self.qrels
        if t == 'dict_of_dict':
            for query_id, docs in self.qrels.items():
                for doc_id, relevance in docs.items():
                    yield Qrel(query_id=query_id, doc_id=doc_id, relevance=relevance)
        if t == 'pd_dataframe':
            if 'iteration' in self.qrels.columns:
                yield from (Qrel(qrel.query_id, qrel.doc_id, qrel.relevance, qrel.iteration) for qrel in self.qrels.itertuples())
            else:
                yield from (Qrel(qrel.query_id, qrel.doc_id, qrel.relevance) for qrel in self.qrels.itertuples())
        if t == 'UNKNOWN':
            raise ValueError(f'unknown qrels format: {err}')

    def as_pd_dataframe(self):
        t, err = self.predict_type()
        if t == 'pd_dataframe':
            if 'iteration' not in self.qrels.columns:
                return self.qrels.assign(iteration=['0'] * len(self.qrels))
            return self.qrels
        else:
            pd = ir_measures.lazylibs.pandas()
            df = pd.DataFrame(self.as_namedtuple_iter())
            if len(df) == 0:
                df = pd.DataFrame([], columns=['query_id', 'doc_id', 'relevance', 'iteration'])
            return df

    @contextlib.contextmanager
    def as_tmp_file(self):
        with tempfile.NamedTemporaryFile(mode='w+t') as f:
            for qrel in self.as_namedtuple_iter():
                f.write('{query_id} 0 {doc_id} {relevance}\n'.format(**qrel._asdict()))
            f.flush()
            f.seek(0)
            yield f




class RunConverter:
    def __init__(self, run, strict=True):
        self.run = run
        self._predicted_format = None
        self.strict = strict # setting strict to false prevents missing columns from raising an error for DFs

    def tee(self, count):
        t, err = self.predict_type()
        if t == 'namedtuple_iter':
            teed_run = itertools.tee(self.run, count)
            return [RunConverter(run) for run in teed_run]
        return [self for _ in range(count)]

    def predict_type(self):
        if self._predicted_format:
            return self._predicted_format
        result = 'UNKNOWN'
        error = None
        if isinstance(self.run, dict):
            result = 'dict_of_dict'
        elif hasattr(self.run, 'itertuples'):
            cols = self.run.columns
            missing_cols = set(ScoredDoc._fields) - set(cols)
            if missing_cols and self.strict:
                error = f'DataFrame missing columns: {list(missing_cols)} (found {list(cols)})'
            else:
                result = 'pd_dataframe'
        elif hasattr(self.run, '__iter__'):
            # peek
            # TODO: is this an OK approach?
            self.run, peek_run = itertools.tee(self.run, 2)
            sentinal = object()
            item = next(peek_run, sentinal)
            if isinstance(item, tuple) and hasattr(item, '_fields'):
                fields = item._fields
                missing_fields = set(ScoredDoc._fields) - set(fields)
                if not missing_fields:
                    result = 'namedtuple_iter'
                else:
                    error = f'namedtuple iter missing fields: {list(missing_fields)} (found {list(fields)})'
            elif item is sentinal:
                result = 'namedtuple_iter'
            else:
                error = 'iterable not a namedtuple iterator'
        else:
            error = f'unexpected format; please provide either: (1) an iterable of namedtuples (fields {ScoredDoc._fields}, e.g., from ir_measures.ScoredDoc); (2) a pandas DataFrame with columns {ScoredDoc._fields}; or (3) a dict-of-dict'
        self._predicted_format = (result, error)
        return result, error

    def as_dict_of_dict(self):
        t, err = self.predict_type()
        if t == 'dict_of_dict':
            return self.run
        else:
            result = {}
            for scored_doc in self.as_namedtuple_iter():
                if scored_doc.query_id not in result:
                    result[scored_doc.query_id] = {}
                result[scored_doc.query_id][scored_doc.doc_id] = scored_doc.score
            return result

    def as_namedtuple_iter(self):
        t, err = self.predict_type()
        if t == 'namedtuple_iter':
            yield from self.run
        if t == 'dict_of_dict':
            for query_id, docs in self.run.items():
                for doc_id, score in docs.items():
                    yield ScoredDoc(query_id=query_id, doc_id=doc_id, score=score)
        if t == 'pd_dataframe':
            yield from (ScoredDoc(sd.query_id, sd.doc_id, sd.score) for sd in self.run.itertuples())
        if t == 'UNKNOWN':
            raise ValueError(f'unknown run format: {err}')

    def as_sorted_namedtuple_iter(self):
        qid = None
        items = []
        def flush():
            return sorted(items, key=lambda x: x.score, reverse=True)
        for item in self.as_namedtuple_iter():
            if qid is None or item.query_id != qid:
                yield from flush()
                items = []
                qid = item.query_id
            items.append(item)
        if qid is not None:
            yield from flush()

    def as_sorteddict(self) -> Dict[str, List[ScoredDoc]]:
        """Returns a map of topic ID -> sorted list of documents"""
        pertopic = defaultdict(list)
        for item in self.as_namedtuple_iter():
            pertopic[item.query_id].append(item)
        return {qid: sorted(items, key=lambda x: x.score, reverse=True) for qid, items in pertopic.items()}

    def as_pd_dataframe(self):
        t, err = self.predict_type()
        if t == 'pd_dataframe':
            return self.run
        else:
            pd = ir_measures.lazylibs.pandas()
            df = pd.DataFrame(self.as_namedtuple_iter())
            if len(df) == 0:
                df = pd.DataFrame([], columns=['query_id', 'doc_id', 'score'])
            return df

    @contextlib.contextmanager
    def as_tmp_file(self):
        with tempfile.NamedTemporaryFile(mode='w+t') as f:
            ranks = {}
            for scoreddoc in self.as_namedtuple_iter():
                key = scoreddoc.query_id
                rank = ranks.setdefault(key, 0)
                f.write('{query_id} Q0 {doc_id} {rank} {score} run\n'.format(**scoreddoc._asdict(), rank=rank))
                ranks[key] += 1
            f.flush()
            f.seek(0)
            yield f


def parse_trec_qrels(file):
    warnings.warn("parse_trec_qrels deprecated in 0.2.0. Please use ir_measures.read_trec_qrels() instead.", DeprecationWarning)
    return read_trec_qrels(file)


def read_trec_qrels(file):
    if hasattr(file, 'read'):
        for line in file:
            if line.strip():
                query_id, iteration, doc_id, relevance = line.split()
                yield Qrel(query_id=query_id, doc_id=doc_id, relevance=int(relevance), iteration=iteration)
    elif isinstance(file, str):
        if '\n' in file:
            yield from read_trec_qrels(io.StringIO(file))
        else:
            reader = gzip.open if file.endswith('.gz') else open
            with reader(file, 'rt') as f:
                yield from read_trec_qrels(f)

def parse_trec_run(file):
    warnings.warn("parse_trec_run deprecated in 0.2.0. Please use ir_measures.read_trec_run() instead.", DeprecationWarning)
    return read_trec_run(file)

def read_trec_run(file):
    if hasattr(file, 'read'):
        for line in file:
            if line.strip():
                query_id, iteration, doc_id, rank, score, tag = line.split()
                yield ScoredDoc(query_id=query_id, doc_id=doc_id, score=float(score))
    elif isinstance(file, str):
        if '\n' in file:
            yield from read_trec_run(io.StringIO(file))
        else:
            reader = gzip.open if file.endswith('.gz') else open
            with reader(file, 'rt') as f:
                yield from read_trec_run(f)


_AST_PARSE_ERROR = 'problem parsing measure {}; must be in format Measure(k1=v1, k2=v2)@c'


def _ast_to_value(node):
    if isinstance(node, ast.Num):
        return node.n
    if isinstance(node, ast.Str):
        return node.s
    if isinstance(node, ast.NameConstant):
        return node.value
    if isinstance(node, ast.Dict):
        return dict(zip(map(_ast_to_value, node.keys), map(_ast_to_value, node.values)))
    raise ValueError(_AST_PARSE_ERROR.format('values must be str, float, int, bool, etc.'))


def parse_measure(measure: str) -> 'ir_measures.Measure':
    node: Union[List[ast.stmt], ast.stmt, ast.expr]
    if isinstance(measure, ir_measures.Measure):
        return measure
    try:
        node = ast.parse(measure).body
    except SyntaxError as e:
        raise ValueError(_AST_PARSE_ERROR.format(str(e)))
    if len(node) != 1:
        raise ValueError(_AST_PARSE_ERROR.format('multiple expressions'))
    node = node[0]
    if not isinstance(node, ast.Expr):
        raise ValueError(_AST_PARSE_ERROR.format('not an expression'))
    node = node.value
    at_param = None
    args = {}
    if isinstance(node, ast.BinOp) and isinstance(node.op, ast.MatMult):
        at_param = _ast_to_value(node.right)
        node = node.left
    if isinstance(node, ast.Call):
        if len(node.args):
            raise ValueError(_AST_PARSE_ERROR.format('args must be named'))
        for keyword in node.keywords:
            args[keyword.arg] = _ast_to_value(keyword.value)
        node = node.func
    if not isinstance(node, ast.Name):
        raise ValueError(_AST_PARSE_ERROR.format('unexpected expression'))
    measure_name = node.id
    if measure_name not in ir_measures.measures.registry:
        raise NameError(f'measure not found: {measure_name}')
    measure_clz = ir_measures.measures.registry[measure_name]
    if at_param is not None:
        args[ measure_clz.AT_PARAM] = at_param
    return  measure_clz(**args)


def parse_trec_measure(measure: str) -> List['ir_measures.Measure']:
    TREC_NAME_MAP = {
        'ndcg': (ir_measures.nDCG, None, None),
        'P': (ir_measures.P, 'cutoff', [5, 10, 15, 20, 30, 100, 200, 500, 1000]),
        'map_cut': (ir_measures.AP, 'cutoff', [5, 10, 15, 20, 30, 100, 200, 500, 1000]),
        'Rndcg': (None, None, None),
        'num_nonrel_judged_ret': (None, None, None),
        'set_map': (ir_measures.SetAP, None, None),
        '11pt_avg': (None, None, None),
        'relative_P': (None, None, None),
        'gm_map': (None, None, None),
        'binG': (None, None, None),
        'set_P': (ir_measures.SetP, None, None),
        'infAP': (ir_measures.infAP, None, None),
        'bpref': (ir_measures.Bpref, None, None),
        'num_rel_ret': (ir_measures.NumRelRet, None, None),
        'ndcg_rel': (None, None, None),
        'recip_rank': (ir_measures.RR, None, None),
        'recall': (ir_measures.R, 'cutoff', [5, 10, 15, 20, 30, 100, 200, 500, 1000]),
        'set_recall': (ir_measures.SetR, None, None),
        'utility': (None, None, None),
        'set_relative_P': (ir_measures.SetRelP, None, None),
        'num_ret': (ir_measures.NumRet, None, None),
        'num_rel': (ir_measures.NumRel, None, None),
        'ndcg_cut': (ir_measures.nDCG, 'cutoff', [5, 10, 15, 20, 30, 100, 200, 500, 1000]),
        # 'runid': (None, None, None), # not really a measure
        'Rprec': (ir_measures.Rprec, None, None),
        'Rprec_mult': (None, None, None),
        'relstring': (None, None, None),
        'gm_bpref': (None, None, None),
        'num_q': (ir_measures.NumQ, None, None),
        'map': (ir_measures.AP, None, None),
        'G': (None, None, None),
        'success': (ir_measures.Success, 'cutoff', [1, 5, 10]),
        'set_F': (ir_measures.SetF, 'beta', [1.]),
        'iprec_at_recall': (ir_measures.IPrec, 'recall', [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]),
    }
    import pytrec_eval
    if measure in pytrec_eval.supported_nicknames:
        result = []
        skipped = []
        for sub_name in sorted(pytrec_eval.supported_nicknames[measure]):
            try:
                result += parse_trec_measure(sub_name)
            except ValueError:
                if sub_name != 'runid':
                    skipped.append(sub_name)
        if skipped:
            print(f'skipped {skipped}: measures not yet supported')
        return result

    # Adapted from <https://github.com/cvangysel/pytrec_eval/blob/master/py/__init__.py#L76>
    RE_BASE = r'{}[\._]([0-9]+(\.[0-9]+)?(,[0-9]+(\.[0-9]+)?)*)'

    # break apart measures in any of the following formats and combine
    #  1) meas          -> {meas: {}}  # either non-parameterized measure or use default params
    #  2) meas.p1       -> {meas: {p1}}
    #  3) meas_p1       -> {meas: {p1}}
    #  4) meas.p1,p2,p3 -> {meas: {p1, p2, p3}}
    #  5) meas_p1,p2,p3 -> {meas: {p1, p2, p3}}
    if measure in TREC_NAME_MAP:
        meas, arg_name, default_args = TREC_NAME_MAP[measure]
        if meas is None:
            raise ValueError(f'known trec name {measure}, but not yet supported')
        if arg_name is not None and default_args is not None:
            result = []
            for arg in default_args:
                result.append(meas(**{arg_name: arg}))
            return result
        return [meas]
    matches = ((m, re.match(RE_BASE.format(re.escape(m)), measure)) for m in TREC_NAME_MAP)
    match = next(filter(lambda x: x[1] is not None, matches), None)
    if match is None:
        raise ValueError('unkonwn measure {}'.format(measure))
    base_meas, meas_args = match[0], match[1].group(1) # type: ignore
    meas, arg_name, _ = TREC_NAME_MAP[base_meas]
    if meas is None:
        raise ValueError(f'known trec name {measure}, but not yet supported')
    if arg_name is None:
        raise ValueError(f'unknown argument for {measure}: {meas_args}')
    result = []
    dtype = meas.SUPPORTED_PARAMS[arg_name].dtype
    for marg in meas_args.split(','):
        result.append(meas(**{arg_name: dtype(marg)}))
    return result


def convert_trec_name(measure: str) -> List['ir_measures.Measure']:
    warnings.warn("convert_trec_name deprecated in 0.2.0. Please use ir_measures.parse_trec_measure() instead.", DeprecationWarning)
    return parse_trec_measure(measure)



from ir_measures import measures


class _RBP(measures.Measure):
    """
    The Rank-Biased Precision (RBP).


    .. cite.dblp:: journals/tois/MoffatZ08
    """
    __name__ = 'RBP'
    NAME = __name__
    PRETTY_NAME = 'Rank-Biased Precision'
    SHORT_DESC = 'A measure of the rate at which utility is gained as a user traverses a result list with a degree of persistence.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'p': measures.ParamInfo(dtype=float, default=0.8, desc='persistence'),
        'rel': measures.ParamInfo(dtype=int, required=False, desc='minimum relevance score to be considered relevant (inclusive), or NOT_PROVIDED to use graded relevance')
    }


RBP = _RBP()
measures.register(RBP)



from ir_measures import measures


class _nDCG(measures.Measure):
    """
    The normalized Discounted Cumulative Gain (nDCG).
    Uses graded labels - systems that put the highest graded documents at the top of the ranking.
    It is normalized wrt. the Ideal NDCG, i.e. documents ranked in descending order of graded label.


    .. cite.dblp:: journals/tois/JarvelinK02
    """
    __name__ = 'nDCG'
    NAME = __name__
    PRETTY_NAME = 'Normalised Discounted Cumulative Gain'
    SHORT_DESC = 'A measure of the total gain a user encounters in a result list, discounted by rank and normalised against an ideal ranking.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'dcg': measures.ParamInfo(dtype=str, choices=['log2', 'exp-log2'], default='log2', desc='DCG formulation'),
        'gains': measures.ParamInfo(dtype=dict, desc='custom gain mapping (int-to-int)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }


nDCG = _nDCG()
NDCG = nDCG
measures.register(nDCG, ['NDCG'])



import sys
import functools
import ir_measures
from ir_measures import providers, measures, Metric
from ir_measures.providers.base import Any, Choices, NOT_PROVIDED


class TrectoolsProvider(providers.Provider):
    """
    trectools

    https://github.com/joaopalotti/trectools


    .. cite.dblp:: conf/sigir/PalottiSZ19
    """
    NAME = 'trectools'
    SUPPORTED_MEASURES = [
        measures._P(cutoff=Any(), rel=Choices(1), judged_only=Choices(False)),
        measures._RR(cutoff=Choices(NOT_PROVIDED), rel=Choices(1), judged_only=Choices(False)),
        measures._Rprec(rel=Choices(1), judged_only=Choices(False)),
        measures._AP(cutoff=Any(), rel=Choices(1), judged_only=Choices(False)),
        measures._nDCG(cutoff=Any(), dcg=Any(), gains=Choices(NOT_PROVIDED), judged_only=Choices(False)),
        measures._Bpref(rel=Choices(1)),
        measures._RBP(cutoff=Any(), p=Any(), rel=Any()),
        # Other supported metrics: urbp, ubpref, alpha_urbp, geometric_map, unjudged
    ]

    def __init__(self):
        super().__init__()
        self.trectools = None

    def _evaluator(self, measures, qrels):
        import pandas as pd
        # Convert qrels to dict_of_dict (input format used by pytrec_eval)
        tmp_qrels = ir_measures.util.QrelsConverter(qrels).as_namedtuple_iter()
        tmp_qrels = pd.DataFrame(tmp_qrels)
        if len(tmp_qrels) == 0:
            tmp_qrels = pd.DataFrame(columns=['query', 'docid', 'rel'], dtype='object')
        else:
            tmp_qrels = tmp_qrels.rename(columns={'query_id': 'query', 'doc_id': 'docid', 'relevance': 'rel'})
        qrels = self.trectools.TrecQrel()
        qrels.qrels_data = tmp_qrels

        invocations = self._build_invocations(measures)

        return TrectoolsEvaluator(measures, qrels, invocations, self.trectools)

    def _build_invocations(self, measures):
        invocations = []
        for measure in measures:
            def depth():
                try:
                    cutoff = measure['cutoff']
                except KeyError:
                    cutoff = NOT_PROVIDED
                if cutoff is NOT_PROVIDED:
                    cutoff = sys.maxsize
                return cutoff
            if measure.NAME == 'P':
                fn = functools.partial(self.trectools.TrecEval.get_precision, depth=depth(), per_query=True, trec_eval=False, removeUnjudged=False)
            elif measure.NAME == 'RR':
                fn = functools.partial(self.trectools.TrecEval.get_reciprocal_rank, depth=depth(), per_query=True, trec_eval=False, removeUnjudged=False)
            elif measure.NAME == 'Rprec':
                fn = functools.partial(self.trectools.TrecEval.get_rprec, depth=depth(), per_query=True, trec_eval=False, removeUnjudged=False)
            elif measure.NAME == 'AP':
                fn = functools.partial(self.trectools.TrecEval.get_map, depth=depth(), per_query=True, trec_eval=False)
            elif measure.NAME == 'nDCG':
                te_mode = {
                    'log2': True,
                    'exp-log2': False
                }[measure['dcg']]
                # trec_eval has other side-effects; namely ordering by score instead of rank.
                # But in our setting, those are always the same so no difference.
                fn = functools.partial(self.trectools.TrecEval.get_ndcg, depth=depth(), per_query=True, trec_eval=te_mode, removeUnjudged=False)
            elif measure.NAME == 'Bpref':
                fn = functools.partial(self.trectools.TrecEval.get_bpref, depth=depth(), per_query=True, trec_eval=False)
            elif measure.NAME == 'RBP':
                rel = measure['rel']
                if rel is not NOT_PROVIDED:
                    # TODO: how to handle different relevance levels? I think the only way is to modify
                    # the dataframe.
                    raise RuntimeError('unsupported')
                else:
                    def fn(ev):
                        return self.trectools.TrecEval.get_rbp(ev, p=measure['p'], depth=depth(), per_query=True, binary_topical_relevance=False, average_ties=True, removeUnjudged=False)[0]
            else:
                raise ValueError(f'unsupported measure {measure}')

            invocations.append((fn, measure))

        return invocations
 
    def initialize(self):
        try:
            import trectools
            self.trectools = trectools
        except ImportError as ex:
            raise RuntimeError('trectools not available', ex)

    def install_instructions(self):
        return 'pip install ir-measures[trectools]'

class TrectoolsEvaluator(providers.Evaluator):
    def __init__(self, measures, qrels, invocations, trectools):
        super().__init__(measures, qrels.qrels_data['query'].unique())
        self.qrels = qrels
        self.invocations = invocations
        self.trectools = trectools

    def _iter_calc(self, run):
        import pandas as pd
        available_qids = set(self.qrels.qrels_data['query'].unique())
        tmp_run = ir_measures.util.RunConverter(run).as_namedtuple_iter()
        tmp_run = pd.DataFrame(tmp_run)
        if len(tmp_run) == 0:
            tmp_run = pd.DataFrame(columns=['query', 'docid', 'score'], dtype='object')
        else:
            tmp_run = tmp_run.rename(columns={'query_id': 'query', 'doc_id': 'docid', 'score': 'score'})
        tmp_run.sort_values(['query', 'score'], ascending=[True, False], inplace=True)
        run = self.trectools.TrecRun()
        run.run_data = tmp_run
        evaluator = self.trectools.TrecEval(run, self.qrels)
        for invocation, measure in self.invocations:
            for query_id, value in invocation(evaluator).itertuples():
                if query_id in available_qids:
                    yield Metric(query_id=query_id, measure=measure, value=value)



providers.register(TrectoolsProvider())



import unittest
import itertools
import ir_measures

class TestMeasures(unittest.TestCase):

    def test_P(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
'''))
        measure = ir_measures.P@5
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.P(rel=2)@5
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.RR
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.RR(rel=2)
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.Rprec
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.Rprec(rel=2)
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.AP
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.AP@2
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.AP(rel=2)
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.AP(rel=2)@2
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.nDCG
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.nDCG@2
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.R@2
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.R(rel=2)@2
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.Bpref
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.Bpref(rel=2)
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.Judged@5
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.Judged@20
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.ERR@2
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.ERR@20
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.nDCG(dcg='exp-log2')@2
        result = list(measure.iter_calc(qrels, run))
        print(result)
        measure = ir_measures.nDCG(dcg='exp-log2')@5
        result = list(measure.iter_calc(qrels, run))
        print(result)
        # measure = ir_measures.RBP(p=[0.5, 0.8, 1.0, 1.2, 1.5])
        # result = list(measure.iter_calc(qrels, run))
        # print(result)
        # measure = ir_measures.RBP(p=[0.5, 0.8, 1.0, 1.2, 1.5])@(1,5,10,20)
        # result = list(measure.iter_calc(qrels, run))
        # print(result)

    def test_measures(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
'''))
        measures = list(itertools.chain.from_iterable([
            [ir_measures.P(rel=r)@k for r in [1, 2] for k in [1,5,10,20,50,100]],
            [ir_measures.R(rel=r)@k for r in [1, 2] for k in [1,5,10,20,50,100]],
            [ir_measures.RR(rel=r) for r in [1, 2]],
            [ir_measures.RR(rel=r)@k for r in [1, 2] for k in [1,5,10,20,50,100]],
            [ir_measures.Rprec(rel=r) for r in [1, 2]],
            [ir_measures.AP(rel=r) for r in [1, 2]],
            [ir_measures.AP(rel=r)@k for r in [1, 2] for k in [1,5,10,20,50,100]],
            [ir_measures.nDCG(dcg=d) for d in ['log2', 'exp-log2']],
            [ir_measures.nDCG(dcg=d)@k for d in ['log2', 'exp-log2'] for k in [1,5,10,20,50,100]],
            [ir_measures.Bpref(rel=r) for r in [1, 2]],
            [ir_measures.Judged@k for k in [1,5,10,20,50,100]],
            [ir_measures.ERR@k for k in [1,5,10,20,50,100]],
            #disable RBP
            #ir_measures.RBP(p=[0.5, 0.8, 1.0, 1.2, 1.5]),
            #ir_measures.RBP(p=[0.5, 0.8, 1.0, 1.2, 1.5])@[1,5,10,20,50,100],
        ]))
        providers = [v for k, v in ir_measures.providers.registry.items() if k != 'trectools']
        for measure in measures:
            values = [(next(p.iter_calc([measure], qrels, run)), p) for p in providers if p.supports(measure)]
            print(measure, len(values))
            for (v1, p1), (v2, p2) in itertools.combinations(values, 2):
                with self.subTest(measure=measure, p1=p1, p2=p2):
                    self.assertAlmostEqual(v1.value, v2.value, places=4, msg=str(measure))



if __name__ == '__main__':
    unittest.main()



from ir_measures import measures


class _INST(measures.Measure):
    """
    INST, a variant of INSQ


    .. cite.dblp:: conf/sigir/BaileyMST15
    """
    __name__ = 'INST'
    NAME = __name__
    PRETTY_NAME = 'INST'
    SHORT_DESC = 'An improved version of INSQ that better handles when either no documents or all retrieved documents are relevant.'
    SUPPORTED_PARAMS = {
        'T': measures.ParamInfo(dtype=float, default=1.0, desc='total desired gain (normalized)'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


class _INSQ(measures.Measure):
    """
    INSQ


    .. cite.dblp:: conf/adcs/MoffatST12
    """
    __name__ = 'INSQ'
    NAME = __name__
    PRETTY_NAME = 'INSQ'
    SHORT_DESC = 'A weighted precision measure based on the conditional probability of the user continuing to the next item.'
    SUPPORTED_PARAMS = {
        'T': measures.ParamInfo(dtype=float, default=1.0, desc='total desired gain (normalized)'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


INST = _INST()
measures.register(INST)

INSQ = _INSQ()
measures.register(INSQ)



from typing import Iterator
import ir_measures
from ir_measures import providers, Metric
from ir_measures.providers.base import Any, NOT_PROVIDED
from ir_measures.measures.accuracy import _Accuracy


class AccuracyEvaluator(providers.Evaluator):
    def __init__(self, measures, qrels, invocations):
        super().__init__(measures, set(qrels.keys()))
        self.qrels = qrels
        self.invocations = invocations

    def iter_calc(self, run) -> Iterator['Metric']:
        """Compute the metrics for the run, discarding topics with no relevant documents"""
        run = ir_measures.util.RunConverter(run).as_sorteddict()

        for measure, cutoff, rel in self.invocations:
            for qid, documents in run.items():
                # Get the relevance assessments 
                qrel = self.qrels.get(qid, {})
                if len(qrel) == 0:
                    continue

                # Count the number of non relevant documents above a
                # relevant one
                _cutoff = cutoff or len(documents)
                nonrels = [0]
                for _, document in zip(range(_cutoff), documents):
                    if qrel.get(document.doc_id, 0) >= rel:
                        nonrels.append(nonrels[-1])
                    else:
                        nonrels[-1] += 1

                # Only report if one relevant document was retrieved
                if len(nonrels) >= 2:
                    value = 1. - sum(nonrels[:-1]) / (float(nonrels[-1]) * (len(nonrels) - 1))
                    yield Metric(query_id=qid, measure=measure, value=value)
 

class AccuracyProvider(providers.Provider):
    """Accuracy provider"""
    NAME = "accuracy"
    SUPPORTED_MEASURES = [
        _Accuracy(cutoff=Any(), rel=Any()),
    ]

    def _evaluator(self, measures, qrels) -> providers.Evaluator:
        invocations = []
        for measure in measures:
            if measure.NAME == _Accuracy.NAME:
                cutoff = 0 if measure['cutoff'] is NOT_PROVIDED else measure['cutoff']
                invocations.append((measure, cutoff, measure['rel']))
            else:
                raise ValueError(f'unsupported measure {measure}')
        qrels = ir_measures.util.QrelsConverter(qrels).as_dict_of_dict()
        
        return AccuracyEvaluator(measures, qrels, invocations)

providers.register(AccuracyProvider())



from ir_measures import measures
from ir_measures.measures.base import SumAgg


class _NumRet(measures.Measure):
    """
    The number of results returned. When rel is provided, counts the number of documents
    returned with at least that relevance score (inclusive).
    """
    __name__ = 'NumRet'
    NAME = __name__
    PRETTY_NAME = 'Number of Retrieved Documents'
    SHORT_DESC = 'The number of documents present in the result set'

    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, required=False, desc='minimum relevance score to be counted (inclusive), or all documents returned if NOT_PROVIDED')
    }

    def aggregator(self):
        return SumAgg()


NumRet = _NumRet()
NumRelRet = NumRet(rel=1)
measures.register(NumRet)
measures.register(NumRelRet, name='NumRelRet')



# These libraries can add a bunch of overhead when imported -- which is bad for command line
# utilities. This file loads them lazily if they are needed.
_cache = {}

def pandas():
    if 'pandas' not in _cache:
        import pandas
        _cache['pandas'] = pandas
    return _cache['pandas']



# flake8: noqa
"""
This module computes evaluation metrics for MSMARCO dataset on the ranking task.
Command line:
python msmarco_eval_ranking.py <path_to_reference_file> <path_to_candidate_file>
Creation Date : 06/12/2018
Last Modified : 1/21/2019
Authors : Daniel Campos <dacamp@microsoft.com>, Rutger van Haasteren <ruvanh@microsoft.com>
Slight modifications by Sean MacAvaney <sean@ir.cs.georgetown.edu> to configure MaxMRRRank, return
results by query, handle edge cases, remove 0 matching check, remove aggregated calculation.
"""
import sys
import statistics

from collections import Counter

MaxMRRRank = 10

def load_reference_from_stream(f):
    """Load Reference reference relevant passages
    Args:f (stream): stream to load.
    Returns:qids_to_relevant_passageids (dict): dictionary mapping from query_id (int) to relevant passages (list of ints). 
    """
    qids_to_relevant_passageids = {}
    for l in f:
        try:
            l = l.strip().split('\t')
            qid = int(l[0])
            if qid in qids_to_relevant_passageids:
                pass
            else:
                qids_to_relevant_passageids[qid] = []
            qids_to_relevant_passageids[qid].append(int(l[2]))
        except:
            raise IOError('\"%s\" is not valid format' % l)
    return qids_to_relevant_passageids

def load_reference(path_to_reference):
    """Load Reference reference relevant passages
    Args:path_to_reference (str): path to a file to load.
    Returns:qids_to_relevant_passageids (dict): dictionary mapping from query_id (int) to relevant passages (list of ints). 
    """
    with open(path_to_reference,'r') as f:
        qids_to_relevant_passageids = load_reference_from_stream(f)
    return qids_to_relevant_passageids

def load_candidate_from_stream(f):
    """Load candidate data from a stream.
    Args:f (stream): stream to load.
    Returns:qid_to_ranked_candidate_passages (dict): dictionary mapping from query_id (int) to a list of 1000 passage ids(int) ranked by relevance and importance
    """
    qid_to_ranked_candidate_passages = {}
    for l in f:
        try:
            l = l.strip().split('\t')
            qid = int(l[0])
            pid = int(l[1])
            rank = int(l[2])
            if qid in qid_to_ranked_candidate_passages:
                pass    
            else:
                # By default, all PIDs in the list of 1000 are 0. Only override those that are given
                tmp = [0] * 1000
                qid_to_ranked_candidate_passages[qid] = tmp
            qid_to_ranked_candidate_passages[qid][rank-1]=pid
        except:
            raise IOError('\"%s\" is not valid format' % l)
    return qid_to_ranked_candidate_passages
                
def load_candidate(path_to_candidate):
    """Load candidate data from a file.
    Args:path_to_candidate (str): path to file to load.
    Returns:qid_to_ranked_candidate_passages (dict): dictionary mapping from query_id (int) to a list of 1000 passage ids(int) ranked by relevance and importance
    """
    
    with open(path_to_candidate,'r') as f:
        qid_to_ranked_candidate_passages = load_candidate_from_stream(f)
    return qid_to_ranked_candidate_passages

def quality_checks_qids(qids_to_relevant_passageids, qids_to_ranked_candidate_passages):
    """Perform quality checks on the dictionaries
    Args:
    p_qids_to_relevant_passageids (dict): dictionary of query-passage mapping
        Dict as read in with load_reference or load_reference_from_stream
    p_qids_to_ranked_candidate_passages (dict): dictionary of query-passage candidates
    Returns:
        bool,str: Boolean whether allowed, message to be shown in case of a problem
    """
    message = ''
    allowed = True

    # Create sets of the QIDs for the submitted and reference queries
    candidate_set = set(qids_to_ranked_candidate_passages.keys())
    ref_set = set(qids_to_relevant_passageids.keys())

    # Check that we do not have multiple passages per query
    for qid in qids_to_ranked_candidate_passages:
        # Remove all zeros from the candidates
        duplicate_pids = set([item for item, count in Counter(qids_to_ranked_candidate_passages[qid]).items() if count > 1])

        if len(duplicate_pids-set([0])) > 0:
            message = "Cannot rank a passage multiple times for a single query. QID={qid}, PID={pid}".format(
                    qid=qid, pid=list(duplicate_pids)[0])
            allowed = False

    return allowed, message

def compute_metrics(qids_to_relevant_passageids, qids_to_ranked_candidate_passages, max_rank=MaxMRRRank):
    """Compute MRR metric
    Args:    
    p_qids_to_relevant_passageids (dict): dictionary of query-passage mapping
        Dict as read in with load_reference or load_reference_from_stream
    p_qids_to_ranked_candidate_passages (dict): dictionary of query-passage candidates
    Returns:
        dict: dictionary of metrics {'MRR': <MRR Score>}
    """
    all_scores = {}
    MRR = 0
    mrr_by_query = {}
    qids_with_relevant_passages = 0
    ranking = []
    for qid in qids_to_ranked_candidate_passages:
        if qid in qids_to_relevant_passageids:
            mrr_by_query[qid] = 0.
            ranking.append(0)
            target_pid = qids_to_relevant_passageids[qid]
            candidate_pid = qids_to_ranked_candidate_passages[qid]
            for i, pid in enumerate(candidate_pid[:max_rank]):
                if pid in target_pid:
                    MRR += 1/(i + 1)
                    mrr_by_query[qid] = 1/(i + 1)
                    ranking.pop()
                    ranking.append(i+1)
                    break
    # if len(ranking) == 0:
    #     raise IOError("No matching QIDs found. Are you sure you are scoring the evaluation set?")

    # MRR = MRR/len(qids_to_relevant_passageids)
    # all_scores[f'MRR @{max_rank}'] = MRR
    all_scores[f'MRR @{max_rank} by query'] = mrr_by_query
    all_scores['QueriesRanked'] = len(qids_to_ranked_candidate_passages)
    return all_scores
                
def compute_metrics_from_files(path_to_reference, path_to_candidate, perform_checks=True, max_rank=MaxMRRRank):
    """Compute MRR metric
    Args:    
    p_path_to_reference_file (str): path to reference file.
        Reference file should contain lines in the following format:
            QUERYID\tPASSAGEID
            Where PASSAGEID is a relevant passage for a query. Note QUERYID can repeat on different lines with different PASSAGEIDs
    p_path_to_candidate_file (str): path to candidate file.
        Candidate file sould contain lines in the following format:
            QUERYID\tPASSAGEID1\tRank
            If a user wishes to use the TREC format please run the script with a -t flag at the end. If this flag is used the expected format is 
            QUERYID\tITER\tDOCNO\tRANK\tSIM\tRUNID 
            Where the values are separated by tabs and ranked in order of relevance 
    Returns:
        dict: dictionary of metrics {'MRR': <MRR Score>}
    """
    
    qids_to_relevant_passageids = load_reference(path_to_reference)
    qids_to_ranked_candidate_passages = load_candidate(path_to_candidate)
    if perform_checks:
        allowed, message = quality_checks_qids(qids_to_relevant_passageids, qids_to_ranked_candidate_passages)
        if message != '': print(message)

    return compute_metrics(qids_to_relevant_passageids, qids_to_ranked_candidate_passages, max_rank)

def main():
    """Command line:
    python msmarco_eval_ranking.py <path_to_reference_file> <path_to_candidate_file>
    """

    if len(sys.argv) == 3:
        path_to_reference = sys.argv[1]
        path_to_candidate = sys.argv[2]
        metrics = compute_metrics_from_files(path_to_reference, path_to_candidate)
        print('#####################')
        for metric in sorted(metrics):
            print('{}: {}'.format(metric, metrics[metric]))
        print('#####################')

    else:
        print('Usage: msmarco_eval_ranking.py <reference ranking> <candidate ranking>')
        exit()
    
if __name__ == '__main__':
    main()



from ir_measures import measures


class _AP(measures.Measure):
    """
    The [Mean] Average Precision ([M]AP). The average precision of a single query is the mean
    of the precision scores at each relevant item returned in a search results list.
    
    AP is typically used for adhoc ranking tasks where getting as many relevant items as possible is. It is commonly referred to as MAP,
    by taking the mean of AP over the query set.

    .. cite.dblp:: journals/ipm/Harman92
    """
    __name__ = 'AP'
    NAME = __name__
    PRETTY_NAME = '(Mean) Average Precision'
    SHORT_DESC = 'The mean of the precision scores at each relevant item retrieved.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }


AP = _AP()
MAP = AP
measures.register(AP, ['MAP'])



from typing import Callable, Optional, Iterable, List, Tuple, TYPE_CHECKING
import ir_measures
from ir_measures import providers, measures, Metric
from ir_measures.measures.base import Measure

if TYPE_CHECKING:
    import pandas as pd


class RuntimeProvider(providers.Provider):
    """
    Supports measures that are defined at runtime via `ir_measures.define()` and
    `ir_measures.define_byquery()`.
    """
    NAME = 'runtime'

    def supports(self, measure):
        measure.validate_params()
        if hasattr(measure, 'runtime_impl'):
            return True
        return False

    def _evaluator(self, measures, qrels):
        # Convert qrels to dict_of_dict (input format used by pytrec_eval)
        qrels = ir_measures.util.QrelsConverter(qrels, strict=False).as_pd_dataframe()
        if "query_id" not in qrels.columns:
            raise ValueError("Required column query_id not found in qrels. Found columns was " + str(qrels.columns.to_list()))
        sort_columns=['query_id']
        if 'doc_id' in qrels.columns:
            sort_columns.append('doc_id')
        qrels = qrels.sort_values(by=sort_columns)
        return RuntimeEvaluator(measures, qrels)

    def run_inputs(self, measures: Iterable[Measure]) -> List[str]:
        """Returns the inputs required by the provided measures in the run.

        .. note::
            This method only supports runtime-defined measures.

        Args:
            measures: A collection of measures to find required inputs for.

        Returns:
            A list of the required inputs.
        """
        inputs = set()
        for measure in measures:
            if hasattr(measure, 'RUN_INPUTS'):
                inputs.update(measure.RUN_INPUTS)
        return list(inputs)

    def qrel_inputs(self, measures: Iterable[Measure]) -> List[str]:
        """Returns the inputs required by the provided measures in the qrels.

        .. note::
            This method only supports runtime-defined measures.

        Args:
            measures: A collection of measures to find required inputs for.

        Returns:
            A list of the required inputs.
        """
        inputs = set()
        for measure in measures:
            if hasattr(measure, 'QREL_INPUTS'):
                inputs.update(measure.QREL_INPUTS)
        return list(inputs)


class RuntimeEvaluator(providers.Evaluator):
    def __init__(self, measures, qrels):
        super().__init__(measures, set(qrels['query_id'].unique()))
        self.qrels = qrels

    def _iter_calc(self, run):
        run = ir_measures.util.RunConverter(run, strict=False).as_pd_dataframe()
        if "query_id" not in run.columns:
            raise ValueError("Required column query_id not found in run. Found columns was " + str(run.columns.to_list()))
        sort_columns = ['query_id']
        sort_orders = [True]
        if 'score' in run.columns:
            sort_columns.append('score')
            sort_orders.append(False)
        run = run.sort_values(by=sort_columns, ascending=sort_orders)
        for measure in self.measures:
            yield from measure.runtime_impl(self.qrels, run)


def define(
    impl: Callable[['pd.DataFrame', 'pd.DataFrame'], Iterable[Tuple[str, float]]],
    name: Optional[str] = None,
    support_cutoff: bool = True,
    *,
    run_inputs: Optional[List[str]] = None,
    qrel_inputs: Optional[List[str]] = None,
    pretty_name : Optional[str] = None,
    short_desc : Optional[str] = None
):
    """Defines a new custom measure from a user-specified function that is is provided all queries at once.

    ``impl`` is a function that accepts (``qrels``, ``run``) and returns an iterable of (qid, score) tuples.

    Most of the time, it is probably easier to use :func:`~ir_measures.define_byquery`, since it operates with one
    query at a time.

    :param impl: A function that takes two pandas DataFrames (qrels and run) and returns an iterable of (qid, score) tuples.
    :param name: The name of the measure (optional)
    :param support_cutoff: Whether the measure supports a cutoff parameter, which reduces the results in run.
    :param run_inputs: Optional list of input columns required by in runs. If not provided, it defaults to ``[query_id, doc_id, score]``.
    :param qrel_inputs: Optional list of input columns required by in qrels. If not provided, it defaults to ``[query_id, doc_id, relevance]``.
    :param pretty_name: Optional str giving a pretty name for the measure.
    :param short_desc: Optional str giving a short description of the measure.  
    """
    _SUPPORTED_PARAMS = {}
    if support_cutoff:
        _SUPPORTED_PARAMS['cutoff'] = measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold')
    if run_inputs is None:
        run_inputs = ['query_id', 'doc_id', 'score']
    if qrel_inputs is None:
        qrel_inputs = ['query_id', 'doc_id', 'relevance']
    class _RuntimeMeasure(measures.Measure):
        nonlocal _SUPPORTED_PARAMS
        nonlocal run_inputs
        nonlocal qrel_inputs
        SUPPORTED_PARAMS = _SUPPORTED_PARAMS
        NAME = name
        __name__ = name
        RUN_INPUTS = run_inputs
        QREL_INPUTS = qrel_inputs
        SHORT_DESC = short_desc
        PRETTY_NAME = pretty_name

        def runtime_impl(self, qrels, run):
            if 'cutoff' in self.params and self.params['cutoff'] is not None:
                cutoff = self.params['cutoff']
                # assumes results already sorted (as is done in RuntimeEvaluator)
                run = run.groupby('query_id').head(cutoff).reset_index(drop=True)
            for qid, score in impl(qrels, run):
                yield Metric(qid, self, score)

        def __repr__(self):
            if self.NAME is not None:
                return super().__repr__()
            return object.__repr__(self)
    return _RuntimeMeasure()


def _byquery_impl(impl):
    def _wrapped(qrels, run):
        for qid, run_subdf in run.groupby("query_id"):
            qrels_subdf = qrels[qrels['query_id'] == qid]
            res = impl(qrels_subdf, run_subdf)
            yield qid, res
    return _wrapped


def define_byquery(
    impl: Callable[['pd.DataFrame', 'pd.DataFrame'], float],
    name: Optional[str] = None,
    support_cutoff: bool = True,
    *,
    run_inputs: Optional[List[str]] = None,
    qrel_inputs: Optional[List[str]] = None,
    pretty_name : Optional[str] = None,
    short_desc : Optional[str] = None
):
    """Defines a new custom measure from a user-specified function that is called once per query.

    ``impl`` is a function that accepts (``qrels``, ``run``) and is called once per query, returning a float
    value each time for the specific query.

    :param impl: A function that takes two pandas DataFrames (qrels and run) and returns a float.
    :param name: The name of the measure (optional)
    :param support_cutoff: Whether the measure supports a cutoff parameter, which reduces the results in run.
    :param run_inputs: Optional list of input columns required by in runs. If not provided, it defaults to ``[query_id, doc_id, score]``.
    :param qrel_inputs: Optional list of input columns required by in qrels. If not provided, it defaults to ``[query_id, doc_id, relevance]``.
    :param pretty_name: Optional str giving a pretty name for the measure.
    :param short_desc: Optional str giving a short description of the measure.
    """
    if name is None:
        if hasattr(impl, '__name__'):
            name = impl.__name__
        else:
            name = repr(impl)
    return define(_byquery_impl(impl), name, support_cutoff, 
                  run_inputs=run_inputs, qrel_inputs=qrel_inputs, 
                  pretty_name=pretty_name, short_desc=short_desc)


providers.register(RuntimeProvider())



import numpy
import unittest
import itertools
import ir_measures
from ir_measures import *
from ir_measures import Metric, CwlMetric
from ir_measures.measures.accuracy import Accuracy


class TestPytrecEval(unittest.TestCase):

    def test_empty(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 2
'''))
        partial_qrels = [q for q in qrels if q.query_id == '0']
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))
        partial_run = [r for r in run if r.query_id == '0']
        empty = []

        # qrels but no run
        self.assertEqual(set(ir_measures.iter_calc([P@5], qrels, empty)), {Metric('0', P@5, 0.), Metric('1', P@5, 0.)})
        self.assertEqual(set(ir_measures.gdeval.iter_calc([ERR@5], qrels, empty)), {Metric('0', ERR@5, 0.), Metric('1', ERR@5, 0.)})
        self.assertEqual(set(ir_measures.judged.iter_calc([Judged@5], qrels, empty)), {Metric('0', Judged@5, 0.), Metric('1', Judged@5, 0.)})
        self.assertEqual(set(ir_measures.msmarco.iter_calc([RR@5], qrels, empty)), {Metric('0', RR@5, 0.), Metric('1', RR@5, 0.)})
        self.assertEqual(set(ir_measures.pytrec_eval.iter_calc([P@5], qrels, empty)), {Metric('0', P@5, 0.), Metric('1', P@5, 0.)})
        self.assertEqual(set(ir_measures.trectools.iter_calc([P@5], qrels, empty)), {Metric('0', P@5, 0.), Metric('1', P@5, 0.)})
        self.assertEqual(set(ir_measures.cwl_eval.iter_calc([P@5], qrels, empty)), {Metric('0', P@5, 0.0), Metric('1', P@5, 0.0)})
        self.assertEqual(set(ir_measures.compat.iter_calc([Compat(p=0.8)], qrels, empty)), {Metric('0', Compat(p=0.8), 0.0), Metric('1', Compat(p=0.8), 0.0)})
        self.assertEqual(set(ir_measures.accuracy.iter_calc([Accuracy()], qrels, empty)), set())

        # qrels but partial run
        self.assertEqual(set(ir_measures.iter_calc([P@5], qrels, partial_run)), {Metric('0', P@5, 0.6), Metric('1', P@5, 0.)})
        self.assertEqual(set(ir_measures.gdeval.iter_calc([ERR@5], qrels, partial_run)), {Metric('0', ERR@5, 0.10175), Metric('1', ERR@5, 0.)})
        self.assertEqual(set(ir_measures.judged.iter_calc([Judged@5], qrels, partial_run)), {Metric('0', Judged@5, 1.), Metric('1', Judged@5, 0.)})
        self.assertEqual(set(ir_measures.msmarco.iter_calc([RR@5], qrels, partial_run)), {Metric('0', RR@5, 0.5), Metric('1', RR@5, 0.)})
        self.assertEqual(set(ir_measures.pytrec_eval.iter_calc([P@5], qrels, partial_run)), {Metric('0', P@5, 0.6), Metric('1', P@5, 0.)})
        self.assertEqual(set(ir_measures.trectools.iter_calc([P@5], qrels, partial_run)), {Metric('0', P@5, 0.6), Metric('1', P@5, 0.)})
        self.assertEqual(set(ir_measures.cwl_eval.iter_calc([P@5], qrels, partial_run)), {CwlMetric('0', P@5, 0.6000000000000001, 3.0, 1.0, 5.0, 5.0), Metric('1', P@5, 0.0)})
        self.assertEqual(set(ir_measures.compat.iter_calc([Compat(p=0.8)], qrels, partial_run)), {Metric('0', Compat(p=0.8), 0.4744431703672816), Metric('1', Compat(p=0.8), 0.0)})
        self.assertEqual(set(ir_measures.accuracy.iter_calc([Accuracy()], qrels, partial_run)), {Metric('0', Accuracy(), 0.5)})

        # run but no qrels
        self.assertEqual(list(ir_measures.iter_calc([P@5], empty, run)), [])
        self.assertEqual(list(ir_measures.gdeval.iter_calc([ERR@5], empty, run)), [])
        self.assertEqual(list(ir_measures.judged.iter_calc([Judged@5], empty, run)), [])
        self.assertEqual(list(ir_measures.msmarco.iter_calc([RR@5], empty, run)), [])
        self.assertEqual(list(ir_measures.pytrec_eval.iter_calc([P@5], empty, run)), [])
        self.assertEqual(list(ir_measures.trectools.iter_calc([P@5], empty, run)), [])
        self.assertEqual(list(ir_measures.cwl_eval.iter_calc([P@5], empty, run)), [])
        self.assertEqual(list(ir_measures.compat.iter_calc([Compat(p=0.8)], empty, run)), [])
        self.assertEqual(list(ir_measures.accuracy.iter_calc([Accuracy()], empty, run)), [])

        # run but partial qrels
        self.assertEqual(set(ir_measures.iter_calc([P@5], partial_qrels, run)), {Metric('0', P@5, 0.6)})
        self.assertEqual(set(ir_measures.gdeval.iter_calc([ERR@5], partial_qrels, run)), {Metric('0', ERR@5, 0.10175)})
        self.assertEqual(set(ir_measures.judged.iter_calc([Judged@5], partial_qrels, run)), {Metric('0', Judged@5, 1.)})
        self.assertEqual(set(ir_measures.msmarco.iter_calc([RR@5], partial_qrels, run)), {Metric('0', RR@5, 0.5)})
        self.assertEqual(set(ir_measures.pytrec_eval.iter_calc([P@5], partial_qrels, run)), {Metric('0', P@5, 0.6)})
        self.assertEqual(set(ir_measures.trectools.iter_calc([P@5], partial_qrels, run)), {Metric('0', P@5, 0.6)})
        self.assertEqual(set(ir_measures.cwl_eval.iter_calc([P@5], partial_qrels, run)), {CwlMetric('0', P@5, 0.6000000000000001, 3.0, 1.0, 5.0, 5.0)})
        self.assertEqual(set(ir_measures.compat.iter_calc([Compat(p=0.8)], partial_qrels, run)), {Metric('0', Compat(p=0.8), 0.4744431703672816)})
        self.assertEqual(set(ir_measures.accuracy.iter_calc([Accuracy()], partial_qrels, run)), {Metric('0', Accuracy(), 0.5)})

        # both no run and no qrels
        self.assertEqual(list(ir_measures.iter_calc([P@5], empty, empty)), [])
        self.assertEqual(list(ir_measures.gdeval.iter_calc([ERR@5], empty, empty)), [])
        self.assertEqual(list(ir_measures.judged.iter_calc([Judged@5], empty, empty)), [])
        self.assertEqual(list(ir_measures.msmarco.iter_calc([RR@5], empty, empty)), [])
        self.assertEqual(list(ir_measures.pytrec_eval.iter_calc([P@5], empty, empty)), [])
        self.assertEqual(list(ir_measures.trectools.iter_calc([P@5], empty, empty)), [])
        self.assertEqual(list(ir_measures.cwl_eval.iter_calc([P@5], empty, empty)), [])
        self.assertEqual(list(ir_measures.compat.iter_calc([Compat(p=0.8)], empty, empty)), [])
        self.assertEqual(list(ir_measures.accuracy.iter_calc([Accuracy()], empty, empty)), [])

        # qrels but no run
        numpy.testing.assert_equal(ir_measures.calc_aggregate([P@5], qrels, empty), {P@5: 0.})
        numpy.testing.assert_equal(ir_measures.gdeval.calc_aggregate([ERR@5], qrels, empty), {ERR@5: 0.})
        numpy.testing.assert_equal(ir_measures.judged.calc_aggregate([Judged@5], qrels, empty), {Judged@5: 0.})
        numpy.testing.assert_equal(ir_measures.msmarco.calc_aggregate([RR@5], qrels, empty), {RR@5: 0.})
        numpy.testing.assert_equal(ir_measures.pytrec_eval.calc_aggregate([P@5], qrels, empty), {P@5: 0.})
        numpy.testing.assert_equal(ir_measures.trectools.calc_aggregate([P@5], qrels, empty), {P@5: 0.})
        numpy.testing.assert_equal(ir_measures.cwl_eval.calc_aggregate([P@5], qrels, empty), {P@5: 0.})
        numpy.testing.assert_equal(ir_measures.compat.calc_aggregate([Compat(p=0.8)], qrels, empty), {Compat(p=0.8): 0.})
        numpy.testing.assert_equal(ir_measures.accuracy.calc_aggregate([Accuracy()], qrels, empty), {Accuracy(): float('NaN')})

        # qrels but partial run
        numpy.testing.assert_equal(ir_measures.calc_aggregate([P@5], qrels, partial_run), {P@5: 0.3})
        numpy.testing.assert_equal(ir_measures.gdeval.calc_aggregate([ERR@5], qrels, partial_run), {ERR@5: 0.050875})
        numpy.testing.assert_equal(ir_measures.judged.calc_aggregate([Judged@5], qrels, partial_run), {Judged@5: 0.5})
        numpy.testing.assert_equal(ir_measures.msmarco.calc_aggregate([RR@5], qrels, partial_run), {RR@5: 0.25})
        numpy.testing.assert_equal(ir_measures.pytrec_eval.calc_aggregate([P@5], qrels, partial_run), {P@5: 0.3})
        numpy.testing.assert_equal(ir_measures.trectools.calc_aggregate([P@5], qrels, partial_run), {P@5: 0.3})
        numpy.testing.assert_equal(ir_measures.cwl_eval.calc_aggregate([P@5], qrels, partial_run), {P@5: 0.30000000000000004})
        numpy.testing.assert_equal(ir_measures.compat.calc_aggregate([Compat(p=0.8)], qrels, partial_run), {Compat(p=0.8): 0.2372215851836408})
        numpy.testing.assert_equal(ir_measures.accuracy.calc_aggregate([Accuracy()], qrels, partial_run), {Accuracy(): 0.5})

        # run but no qrels
        numpy.testing.assert_equal(ir_measures.calc_aggregate([P@5], empty, run), {P@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.gdeval.calc_aggregate([ERR@5], empty, run), {ERR@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.judged.calc_aggregate([Judged@5], empty, run), {Judged@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.msmarco.calc_aggregate([RR@5], empty, run), {RR@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.pytrec_eval.calc_aggregate([P@5], empty, run), {P@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.trectools.calc_aggregate([P@5], empty, run), {P@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.cwl_eval.calc_aggregate([P@5], empty, run), {P@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.compat.calc_aggregate([Compat(p=0.8)], empty, run), {Compat(p=0.8): float('NaN')})
        numpy.testing.assert_equal(ir_measures.accuracy.calc_aggregate([Accuracy()], empty, run), {Accuracy(): float('NaN')})

        # run but partial qrels
        numpy.testing.assert_equal(ir_measures.calc_aggregate([P@5], partial_qrels, run), {P@5: 0.6})
        numpy.testing.assert_equal(ir_measures.gdeval.calc_aggregate([ERR@5], partial_qrels, run), {ERR@5: 0.10175})
        numpy.testing.assert_equal(ir_measures.judged.calc_aggregate([Judged@5], partial_qrels, run), {Judged@5: 1.0})
        numpy.testing.assert_equal(ir_measures.msmarco.calc_aggregate([RR@5], partial_qrels, run), {RR@5: 0.5})
        numpy.testing.assert_equal(ir_measures.pytrec_eval.calc_aggregate([P@5], partial_qrels, run), {P@5: 0.6})
        numpy.testing.assert_equal(ir_measures.trectools.calc_aggregate([P@5], partial_qrels, run), {P@5: 0.6})
        numpy.testing.assert_equal(ir_measures.cwl_eval.calc_aggregate([P@5], partial_qrels, run), {P@5: 0.6000000000000001})
        numpy.testing.assert_equal(ir_measures.compat.calc_aggregate([Compat(p=0.8)], partial_qrels, run), {Compat(p=0.8): 0.4744431703672816})
        numpy.testing.assert_equal(ir_measures.accuracy.calc_aggregate([Accuracy()], partial_qrels, run), {Accuracy(): 0.5})

        # both no run and no qrels
        numpy.testing.assert_equal(ir_measures.calc_aggregate([P@5], empty, empty), {P@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.gdeval.calc_aggregate([ERR@5], empty, empty), {ERR@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.judged.calc_aggregate([Judged@5], empty, empty), {Judged@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.msmarco.calc_aggregate([RR@5], empty, empty), {RR@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.pytrec_eval.calc_aggregate([P@5], empty, empty), {P@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.trectools.calc_aggregate([P@5], empty, empty), {P@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.cwl_eval.calc_aggregate([P@5], empty, empty), {P@5: float('NaN')})
        numpy.testing.assert_equal(ir_measures.compat.calc_aggregate([Compat(p=0.8)], empty, empty), {Compat(p=0.8): float('NaN')})
        numpy.testing.assert_equal(ir_measures.accuracy.calc_aggregate([Accuracy()], empty, empty), {Accuracy(): float('NaN')})



if __name__ == '__main__':
    unittest.main()



from ir_measures.measures.base import Measure, BaseMeasure, ParamInfo, MeanAgg, SumAgg, registry, register
from ir_measures.measures.accuracy import Accuracy, _Accuracy
from ir_measures.measures.ap import AP, MAP, _AP
from ir_measures.measures.bpm import BPM, _BPM
from ir_measures.measures.bpref import Bpref, BPref, _Bpref
from ir_measures.measures.compat import Compat, _Compat
from ir_measures.measures.diversity import ERR_IA, _ERR_IA, nERR_IA, _nERR_IA, alpha_DCG, α_DCG, _alpha_DCG, alpha_nDCG, α_nDCG, _alpha_nDCG, NRBP, _NRBP, nNRBP, _nNRBP, AP_IA, MAP_IA, _AP_IA, P_IA, _P_IA, StRecall, _StRecall
from ir_measures.measures.err import ERR, _ERR
from ir_measures.measures.inst import INST, _INST, INSQ, _INSQ
from ir_measures.measures.infap import infAP, _infAP
from ir_measures.measures.iprec import IPrec, _IPrec
from ir_measures.measures.judged import Judged, _Judged
from ir_measures.measures.ndcg import nDCG, NDCG, _nDCG
from ir_measures.measures.nerr import NERR8, NERR9, NERR10, NERR11, _NERR8, _NERR9, _NERR10, _NERR11
from ir_measures.measures.numq import NumQ, _NumQ
from ir_measures.measures.numrel import NumRel, _NumRel
from ir_measures.measures.numret import NumRet, NumRelRet, _NumRet
from ir_measures.measures.p import P, Precision, _P
from ir_measures.measures.r import R, Recall, _R
from ir_measures.measures.rbp import RBP, _RBP
from ir_measures.measures.rprec import Rprec, RPrec, _Rprec
from ir_measures.measures.rr import RR, MRR, _RR
from ir_measures.measures.sdcg import SDCG, _SDCG
from ir_measures.measures.set_measures import SetP, SetRelP, _SetP, SetR, _SetR, SetF, _SetF, SetAP, _SetAP
from ir_measures.measures.success import Success, _Success

# include all (to make ruff happy)
__all__ = [
    'Measure', 'BaseMeasure', 'ParamInfo', 'MeanAgg', 'SumAgg', 'registry', 'register',
    'Accuracy', '_Accuracy', 'AP', 'MAP', '_AP', 'BPM', '_BPM', 'Bpref', 'BPref', '_Bpref', 'Compat', '_Compat',
    'ERR', '_ERR', 'ERR_IA', '_ERR_IA', 'nERR_IA', '_nERR_IA', 'alpha_DCG', 'α_DCG', '_alpha_DCG', 'alpha_nDCG', 'α_nDCG', '_alpha_nDCG',
    'NRBP', '_NRBP', 'nNRBP', '_nNRBP', 'AP_IA', 'MAP_IA', '_AP_IA', 'P_IA', '_P_IA', 'StRecall', '_StRecall',
    'INST', '_INST', 'INSQ', '_INSQ', 'infAP', '_infAP', 'IPrec', '_IPrec', 'Judged', '_Judged', 'nDCG', 'NDCG', '_nDCG',
    'NERR8', 'NERR9', 'NERR10', 'NERR11', '_NERR8', '_NERR9', '_NERR10', '_NERR11', 'NumQ', '_NumQ', 'NumRel', '_NumRel',
    'NumRet', 'NumRelRet', '_NumRet', 'P', 'Precision', '_P', 'R', 'Recall', '_R', 'RBP', '_RBP', 'Rprec', 'RPrec', '_Rprec',
    'RR', 'MRR', '_RR', 'SDCG', '_SDCG', 'SetP', 'SetRelP', '_SetP', 'SetR', '_SetR', 'SetF', '_SetF', 'SetAP', '_SetAP',
    'Success', '_Success'
]

# enable "from ir_measures.measures import *" --- on purpuse, do not include _-prefixed versions,
# as these are intended for internal use
__all__ = [
    'Measure', 'BaseMeasure', 'ParamInfo', 'MeanAgg', 'SumAgg', 'registry', 'register',
    'Accuracy', 'AP', 'MAP', 'BPM', 'Bpref', 'BPref', 'Compat', 'ERR_IA', 'nERR_IA', 'alpha_DCG', 'α_DCG', 'alpha_nDCG', 'α_nDCG', 'NRBP', 'nNRBP', 'AP_IA', 'MAP_IA', 'P_IA', 'StRecall', 'ERR', 'INST', 'INSQ', 'infAP', 'IPrec', 'Judged', 'nDCG', 'NDCG', 'NERR8', 'NERR9', 'NERR10', 'NERR11', 'NumQ', 'NumRel', 'NumRet', 'NumRelRet', 'P', 'Precision', 'R', 'Recall', 'RBP', 'Rprec', 'RPrec', 'RR', 'MRR', 'SDCG', 'SetP', 'SetRelP', 'SetR', 'SetF', 'SetAP', 'Success'
]



__version__ = "0.4.3"
import sys
import logging
from ir_measures import util
from ir_measures import lazylibs
from ir_measures.util import (
    parse_measure, parse_trec_measure, read_trec_qrels, read_trec_run,
    Qrel, ScoredDoc, Metric, CalcResults,
    GenericQrel, # deprecated; replaced with Qrel
    GenericScoredDoc, # deprecated; replaced with ScoredDoc
    convert_trec_name, # deprecated; replaced with parse_trec_measure
    parse_trec_qrels, # deprecated; replaced with read_trec_qrels
    parse_trec_run, # deprecated; replaced with read_trec_run
)
from ir_measures import measures
from ir_measures.measures import (
    Accuracy, alpha_DCG, alpha_nDCG, AP, AP_IA, BaseMeasure, BPM, Bpref, BPref, Compat, ERR, ERR_IA, infAP, INSQ, INST,
    IPrec, Judged, MAP, MAP_IA, MeanAgg, Measure, MRR, nDCG, NDCG, NERR10, NERR11, NERR8, NERR9, nERR_IA, nNRBP, NRBP,
    NumQ, NumRel, NumRelRet, NumRet, P, P_IA, ParamInfo, Precision, R, RBP, Recall, Rprec, RPrec, RR, SDCG, SetAP, SetF,
    SetP, SetR, SetRelP, StRecall, Success, SumAgg, α_DCG, α_nDCG
)
from ir_measures import providers
from ir_measures.providers import Provider, Evaluator



logger = logging.getLogger('ir_measures')
logger.setLevel('WARNING')
log_handler = logging.StreamHandler(sys.stderr)
log_handler.setFormatter(logging.Formatter('[%(name)s] [%(levelname)s] %(message)s'))
logger.addHandler(log_handler)

# providers
accuracy = providers.registry['accuracy']
cwl_eval = providers.registry['cwl_eval']
compat = providers.registry['compat']
gdeval = providers.registry['gdeval']
pytrec_eval = providers.registry['pytrec_eval']
trectools = providers.registry['trectools']
judged = providers.registry['judged']
msmarco = providers.registry['msmarco']
pyndeval = providers.registry['pyndeval']
ranx = providers.registry['ranx']
runtime = providers.registry['runtime']

define = providers.define
define_byquery = providers.define_byquery

CwlMetric = providers.CwlMetric

DefaultPipeline = providers.FallbackProvider([
    runtime,
    pytrec_eval,
    cwl_eval,
    compat,
    pyndeval,
    # trectools,  # buggy; will add back later
    judged,
    msmarco,
    gdeval,  # doesn't work when installed from package #9
    accuracy,
    ranx,
])
evaluator = DefaultPipeline.evaluator
calc_ctxt = DefaultPipeline.calc_ctxt # deprecated; replaced with evaluator
iter_calc = DefaultPipeline.iter_calc
calc_aggregate = DefaultPipeline.calc_aggregate
calc = DefaultPipeline.calc
run_inputs = DefaultPipeline.run_inputs
qrel_inputs = DefaultPipeline.qrel_inputs

__all__ = [
    'accuracy', 'cwl_eval', 'compat', 'gdeval', 'pytrec_eval', 'trectools', 'judged', 'msmarco', 'pyndeval', 'ranx',
    'runtime',
    'define', 'define_byquery',
    'CwlMetric',
    'DefaultPipeline',
    'evaluator', 'calc_ctxt', 'iter_calc', 'calc_aggregate', 'calc',
    'Accuracy', 'alpha_DCG', 'alpha_nDCG', 'AP', 'AP_IA', 'BaseMeasure', 'BPM', 'Bpref', 'BPref', 'Compat', 'ERR',
    'ERR_IA', 'infAP', 'INSQ', 'INST', 'IPrec', 'Judged', 'MAP', 'MAP_IA', 'MeanAgg', 'Measure', 'MRR', 'nDCG', 'NDCG',
    'NERR10', 'NERR11', 'NERR8', 'NERR9', 'nERR_IA', 'nNRBP', 'NRBP', 'NumQ', 'NumRel', 'NumRelRet', 'NumRet', 'P',
    'P_IA', 'ParamInfo', 'Precision', 'R', 'RBP', 'Recall', 'Rprec', 'RPrec', 'RR', 'SDCG', 'SetAP', 'SetF', 'SetP',
    'SetR', 'SetRelP', 'StRecall', 'Success', 'SumAgg', 'α_DCG', 'α_nDCG',
    'Measure',
    'Provider', 'Evaluator',
    'Qrel', 'ScoredDoc', 'Metric', 'CalcResults',
    'GenericQrel', 'GenericScoredDoc',
    'convert_trec_name', 'parse_trec_measure',
    'read_trec_qrels', 'read_trec_run',
    'parse_measure', 'parse_trec_qrels', 'parse_trec_run',
    'util',
    'measures',
    'providers',
    'lazylibs',
    '__version__',
]



import os
import sys
import argparse
import json
import ir_measures
from ir_measures.util import Metric

DEFAULT_PLACES = 4
SUMMARY_QID = 'all'


def tsv_output(args):
    if args.places == -1:
        place_format = '' # --places -1 indicates no format string (show all places, exp notation, etc.)
    else:
        place_format = f'.{args.places}f'
    def wrapped(result):
        if args.by_query:
            print(f'{result.query_id}\t{result.measure}\t{result.value:{place_format}}')
        else:
            print(f'{result.measure}\t{result.value:{place_format}}')
    return wrapped


def jsonl_output(args):
    if args.places != DEFAULT_PLACES:
        sys.stderr.write('--places is ignored when using --output jsonl\n')
    def wrapped(result):
        result = result._replace(measure=str(result.measure))._asdict()
        if not args.by_query:
            del result['query_id']
        print(json.dumps(result))
    return wrapped


OUTPUT_FORMATS = {
    'tsv': tsv_output,
    'jsonl': jsonl_output,
}


def main_cli():
    parser = argparse.ArgumentParser()
    parser.add_argument('qrels')
    parser.add_argument('run')
    parser.add_argument('measures', nargs='+')
    parser.add_argument('--places', '-p', type=int, default=DEFAULT_PLACES)
    parser.add_argument('--by_query', '-q', action='store_true')
    parser.add_argument('--no_summary', '-n', action='store_true')
    parser.add_argument('--output_format', '-o', choices=OUTPUT_FORMATS.keys(), default='tsv')
    parser.add_argument('--provider', choices=ir_measures.providers.registry.keys())
    args = parser.parse_args()
    qrels = _get_qrels(args)
    run = ir_measures.read_trec_run(args.run)
    measures = _get_measures(args)
    calc_obj = ir_measures
    if args.provider:
        calc_obj = ir_measures.providers.registry[args.provider]
    output = OUTPUT_FORMATS[args.output_format](args)
    if args.by_query:
        aggs = {m: m.aggregator() for m in measures} if not args.no_summary else None
        for result in calc_obj.iter_calc(measures, qrels, run):
            output(result)
            if aggs:
                aggs[result.measure].add(result.value)
        if aggs:
            for measure in measures:
                output(Metric(query_id=SUMMARY_QID, measure=measure, value=aggs[measure].result()))
    else:
        assert not args.no_summary, "--no_summary (-n) only supported with --by_query (-q)"
        results = calc_obj.calc_aggregate(measures, qrels, run)
        for measure in measures:
            output(Metric(query_id=SUMMARY_QID, measure=measure, value=results[measure]))


def _get_qrels(args):
    # gets the qrels, either from a file (priority) or from ir_datasets (if installed)
    if os.path.exists(args.qrels):
        return ir_measures.read_trec_qrels(args.qrels)
    irds_available = False
    try:
        import ir_datasets
        irds_available = True
    except ImportError:
        sys.stderr.write('Skipping ir_datasets lookup. To use this feature, install ir_datasets.\n')
    if irds_available:
        try:
            ds = ir_datasets.load(args.qrels)
            if ds.has_qrels():
                return ds.qrels_iter()
            sys.stderr.write(f'ir_datasets ID {args.qrels} found but does not provide qrels.\n')
            sys.exit(-1)
        except KeyError:
            sys.stderr.write(f'{args.qrels} not found. (checked file and ir_datasets)\n')
            sys.exit(-1)
    sys.stderr.write(f'{args.qrels} not found.\n')
    sys.exit(-1)


def _get_measures(args):
    measures, errors = [], []
    for mstr in args.measures:
        for m in mstr.split():
            try:
                measure = ir_measures.parse_measure(m)
                if measure not in measures:
                    measures.append(measure)
            except ValueError:
                errors.append(f'syntax error: {m}')
            except NameError:
                errors.append(f'unknown measure: {m}')
    if errors:
        sys.stderr.write('\n'.join(['error parsing measures'] + errors + ['']))
        sys.exit(-1)
    return measures


if __name__ == '__main__':
    try:
        main_cli()
    except BrokenPipeError:
        sys.stderr.write('BrokenPipe\n')
    except KeyboardInterrupt:
        sys.stderr.write('KeyboardInterrupt\n')



from ir_measures import measures


class _SDCG(measures.Measure):
    """
    The Scaled Discounted Cumulative Gain (SDCG), a variant of nDCG that assumes more
    fully-relevant documents exist but are not labeled.
    """
    __name__ = 'SDCG'
    NAME = __name__
    PRETTY_NAME = 'Scaled Discounted Cumulative Gain'
    SHORT_DESC = 'nDCG, but assuming there are at least "cutoff" relevant documents'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'dcg': measures.ParamInfo(dtype=str, choices=['log2'], default='log2', desc='DCG formulation'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


SDCG = _SDCG()
measures.register(SDCG)



import unittest
import ir_measures
from ir_measures.measures import Judged


class TestMeasures(unittest.TestCase):

    def test_judged(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 2
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.judged

        expected_results = [
            [Judged, [('0', 1.0), ('1', 1.0 / 4.0)]],
            [Judged(cutoff=1000), [('0', 1.0), ('1', 1.0 / 4.0)]],
            [Judged(cutoff=10), [('0', 1.0), ('1', 1.0 / 4.0)]],
            [Judged(cutoff=3), [('0', 1.0), ('1', 1.0 / 3.0)]],
            [Judged(cutoff=1), [('0', 1.0), ('1', 0.0)]],
        ]
        for measure, expected in expected_results:
            with self.subTest(measure=measure):
                self.assertTrue(provider.supports(measure))
                results = list(provider.iter_calc([measure], qrels, run))
                for result, (query_id, value) in zip(results, expected):
                    self.assertAlmostEqual(result.query_id, query_id, delta=0.0001)
                    self.assertAlmostEqual(result.value, value, delta=0.0001)


if __name__ == '__main__':
    unittest.main()



from ir_measures import measures


class _Compat(measures.Measure):
    """
    Compatibility measure desribed in:


    .. cite.dblp:: journals/tois/ClarkeVS21
    """
    __name__ = 'Compat'
    NAME = __name__
    PRETTY_NAME = 'Compatibility'
    SHORT_DESC = 'The Rank Biased Overlap between the results and an ideal ranking.'
    SUPPORTED_PARAMS = {
        'p': measures.ParamInfo(dtype=float, default=0.95, desc='persistence'),
        'normalize': measures.ParamInfo(dtype=bool, default=True, desc='apply normalization for finite ideal rankings'),
    }


Compat = _Compat()
measures.register(Compat)



from ir_measures import measures


class _RR(measures.Measure):
    """
    The [Mean] Reciprocal Rank ([M]RR) is a precision-focused measure that scores based on the reciprocal of the rank of the
    highest-scoring relevance document. An optional cutoff can be provided to limit the
    depth explored. rel (default 1) controls which relevance level is considered relevant.


    .. cite.dblp:: journals/ir/KantorV00
    """
    __name__ = 'RR'
    NAME = __name__
    PRETTY_NAME = '(Mean) Reciprocal Rank'
    SHORT_DESC = 'The reciprocal of the rank of the first relevant document.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }


RR = _RR()
MRR = RR
measures.register(RR, ['MRR'])



from ir_measures import measures


class _BPM(measures.Measure):
    """
    The Bejeweled Player Model (BPM).

    .. cite.dblp:: conf/sigir/ZhangLLZXM17
    """
    __name__ = 'BPM'
    NAME = __name__
    PRETTY_NAME = 'Bejeweled Player Model'
    SHORT_DESC = 'A measure that balances both gain and user patience to determine when they stop traversing search results.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'T': measures.ParamInfo(dtype=float, default=1., desc='total desired gain (normalized)'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


BPM = _BPM()
measures.register(BPM)



import unittest
import itertools
import ir_measures
from ir_measures import *


class TestPynedvalEval(unittest.TestCase):

    def test_measures(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 1 D2 1
0 1 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 1 D5 2
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))
        measures = [
            ERR_IA@5,
            ERR_IA(rel=2)@10,
            nERR_IA@5,
            nERR_IA(rel=2)@10,
            alpha_DCG@5,
            alpha_nDCG@5,
            NRBP,
            NRBP(rel=2),
            nNRBP,
            nNRBP(rel=2),
            AP_IA,
            AP_IA(rel=2),
            P_IA@5,
            P_IA(rel=2)@10,
            StRecall@5,
            StRecall(rel=2)@10,
        ]
        ir_measures.pyndeval.calc_aggregate(measures, qrels, run)

    def test_ERR_IA(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 1 D2 1
0 1 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 1 D5 2
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.pyndeval
        measure = ir_measures.ERR_IA@20
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertAlmostEqual(result[0].value, 0.4659, places=4)
        self.assertEqual(result[1].query_id, "1")
        self.assertAlmostEqual(result[1].value, 0.1803, places=4)
        self.assertAlmostEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.3231, places=4)



if __name__ == '__main__':
    unittest.main()
